/* 
 * Mundu@Attention
 * Enable Attention Manager alerts for Mundu IM
 * copyright (c) 2008, daniel grobe sachs <dgsachs@nekito.net>
 *
 * This is free software. See included LICENSE file for details.
 */

#include <PalmOS.h>
#include <palmOne_68K.h>
#include <68K/System/HsExt.h>
#include <Common/System/HsExtCommon.h>

#include "resource.h"

#define STARTUP_CODE_SIZE	152		// size of prc-tools crt0.c

UInt32 PilotMain(UInt16 launchCode, MemPtr cmdPBP, UInt16 launchFlags);

// Endian swapper
#define EndianSwap16(n) (((((UInt16) n) << 8) & 0xFF00) | \
		         ((((UInt16) n) >> 8) & 0x00FF))

static void RegisterForNotifications();
static void DeregisterNotifications();

#define FTR_PTR 		1
#define FTR_ATTN_PTR		2
#define FTR_ICON_PTR		3
#define POPUPTIME_FTR		100
#define IM_COUNT_FTR		101
#define IM_USER_PTR_FTR		102
#define IM_MESG_PTR_FTR		103
#define MUNDU_CARDNO_FTR 	104
#define MUNDU_DBID_FTR 		105
#define MY_CARDNO_FTR		106
#define MY_DBID_FTR		107

#define POPUP_CLEAR_TIME 15 // seconds

#define MUNDU_MAIN_WINDOW 	3100
#define MUNDU_CHAT_WINDOW	2000
#define MUNDU_BG_WINDOW 	4300
#define MUNDU_BG_HIDE      	4304
#define MUNDU_PRESENCE_WINDOW	4950
#define MUNDU_PRESENCE_HIDE	4951
#define MUNDU_CHAT_BITMAP	24900

void IMInitNotifications(void);
void IMAddNotification(Char *from, Char *msg);
void IMResetNotifications(void);
void IMDrawIcon(int x, int y);

// Notification handler
static Err HandleNotification(SysNotifyParamType *np)
{
	UInt16 cardNo;
	LocalID dbID;
	UInt32 munduCard, munduDb;

        UInt32 time;
	UInt32 clear_time = 0;
	UInt16 frm;
        //struct EventType newevent;
        EventType *eventP = (EventType*)np->notifyDetailsP;

        // Check if we're deleting the application
        if( np->notifyType == sysNotifyDeleteProtectedEvent )
        {
                if( ((SysNotifyDBInfoType *)np->notifyDetailsP)->creator == CREATOR )
                {
                        DeregisterNotifications();
                        np->handled = true;
                }

                return 0;
        }

        // Is this a queue event?
        if( np->notifyType != sysNotifyEventDequeuedEvent )
                return 0;

	// Check application ID
	SysCurAppDatabase( &cardNo, &dbID );

	FtrGet(CREATOR,MUNDU_CARDNO_FTR,&munduCard);
	FtrGet(CREATOR,MUNDU_DBID_FTR,&munduDb);

#if 0
	{
		UInt32 t;
		Char *pt;

		Char p[20];

		FtrGet(CREATOR,IM_COUNT_FTR,&t);
		StrIToA(p,t);
		WinDrawChars(p,StrLen(p),0,0);

		pt = NULL;
		FtrGet(CREATOR,IM_USER_PTR_FTR,&pt);
		if( pt ) 
			WinDrawChars(pt,StrLen(pt),0,16);
		
		pt = NULL;
		FtrGet(CREATOR,IM_MESG_PTR_FTR,&pt);
		if( pt ) 
			WinDrawChars(pt,StrLen(pt),0,32);
	}
#endif
	
	// If it's not Mundu, reset the timer and exit
	if( (cardNo != munduCard) || (dbID != munduDb) )
	{
		FtrSet(CREATOR,POPUPTIME_FTR,0);
		return 0;
	}

	// At this point we know it's Mundu. Get the Window ID.

	frm = FrmGetActiveFormID();
	FtrGet(CREATOR,POPUPTIME_FTR,&clear_time);
	time = TimGetTicks();

	// In main application?
	if( (frm == MUNDU_MAIN_WINDOW) || (frm == MUNDU_CHAT_WINDOW) )
	{
		IMResetNotifications();
	}
		
	// User activity (temporarily) overrides clear
	if( ((frm != MUNDU_BG_WINDOW) && (frm != MUNDU_PRESENCE_WINDOW)) || 
	    (eventP->eType == EndianSwap16(penDownEvent)) ||
	    (eventP->eType == EndianSwap16(penMoveEvent)) ||
	    (eventP->eType == EndianSwap16(penUpEvent))   ||
	    (eventP->eType == EndianSwap16(keyDownEvent)) )
	{
		// Reset timer and exit
		FtrSet(CREATOR,POPUPTIME_FTR,0);
		return 0;
	}

#if 0
	// Spit out debugging data
	{
		FormType *frmp;
		Char *user, *mesg;

		// Retrieve strings
		frmp = FrmGetActiveForm();
		user = FldGetTextPtr(FrmGetObjectPtr(frmp, FrmGetObjectIndex(frmp, 4308)));
		mesg = FldGetTextPtr(FrmGetObjectPtr(frmp, FrmGetObjectIndex(frmp, 4306)));

		WinDrawChars(user,StrLen(user),0,0);
		WinDrawChars(mesg,StrLen(mesg),0,32);
	}
#endif

	// If clear time isn't set, set it. 
	if( !clear_time )
	{
		FtrSet(CREATOR,POPUPTIME_FTR,
			time+(POPUP_CLEAR_TIME * SysTicksPerSecond()));
	}
			
	// If clear time is elapsed, clear the dialog and cue Attention Manager
	else if( clear_time && (time > clear_time) )
	{
		FormType *frmp;
		char buf[128];
		Char *user, *mesg;

		EventType *newEvent;
		MemSet(buf,128,0);

		// If this is a background chat window, cue attention manager
		if( frm == MUNDU_BG_WINDOW )
		{
			// Retrieve strings
			frmp = FrmGetActiveForm();
			user = FldGetTextPtr(FrmGetObjectPtr(frmp, FrmGetObjectIndex(frmp, 4308)));
			mesg = FldGetTextPtr(FrmGetObjectPtr(frmp, FrmGetObjectIndex(frmp, 4306)));

                	//SndPlaySystemSound(sndInfo);

			// Add IM to Attention Manager
			IMAddNotification(user, mesg);
		
			// Clear the window
			newEvent = (EventType *)buf;
			newEvent->eType = ctlSelectEvent;
			newEvent->data.ctlSelect.controlID = MUNDU_BG_HIDE;
			EvtAddEventToQueue(newEvent);
		}
		else	// just clear the window
		{
			// Clear the window
			newEvent = (EventType *)buf;
			newEvent->eType = ctlSelectEvent;
			newEvent->data.ctlSelect.controlID = MUNDU_PRESENCE_HIDE;
			EvtAddEventToQueue(newEvent);
		}

		// Reset timer
		FtrSet(CREATOR,POPUPTIME_FTR,0);
	}

	// Else waiting for clear time to elapse
	else if( clear_time )
	{
		// Make sure we have events to process
		EvtSetNullEventTick(time + SysTicksPerSecond());
	}

	return 0;
}

static Err HandleAttentionMgr(AttnCommandType command, UInt32 userData,
					 AttnCommandArgsType *commandArgsP)
{
	UInt32 munduCard, munduDb;

	Char *user_data = NULL;
	Char *mesg_data = NULL;
	UInt32 counter;

	Char msg[80] = "";

	Char StrMessage[] = "1 IM";
	Char StrMessages[] = " IMs";
	Char StrFrom[] = " from ";
	Char StrColon[] = ":";

	// Get counter
	FtrGet(CREATOR,IM_COUNT_FTR,&counter);

	// Get message data
	FtrGet(CREATOR,IM_USER_PTR_FTR,&user_data);
	FtrGet(CREATOR,IM_MESG_PTR_FTR,&mesg_data);

	if( counter == 1 )
	{
		StrNCopy(msg,StrMessage,80);
	}
	else
	{
		StrIToA(msg,counter);
		StrNCat(msg,StrMessages,80);
	}

	if( user_data )
	{
		StrNCat(msg,StrFrom,80);
		StrNCat(msg,user_data,80);
	}

	if( mesg_data )
	{
		StrNCat(msg,StrColon,80);
	}

	if ( kAttnCommandDrawDetail == command )
	{
		Coord x = commandArgsP->drawDetail.bounds.topLeft.x;
		Coord y = commandArgsP->drawDetail.bounds.topLeft.y;
		Coord w = commandArgsP->drawDetail.bounds.extent.x;

		FontID curFont = FntSetFont (stdFont);

		IMDrawIcon(x + ((w/2) - 6), y + 3);
		
		WinDrawTruncChars(msg, StrLen(msg),
						  x,
						  y + 10, 
						  w);

		if( mesg_data )
		{
			WinDrawTruncChars(mesg_data, StrLen(mesg_data),
						  x,
						  y + 21, 
						  w);
		}
		FntSetFont (curFont);
	}
	else if ( kAttnCommandDrawList == command )
	{
		Coord x = commandArgsP->drawList.bounds.topLeft.x;
		Coord y = commandArgsP->drawList.bounds.topLeft.y;
		Coord w = commandArgsP->drawList.bounds.extent.x;

		FontID curFont = FntSetFont (stdFont);

		IMDrawIcon(x+1, y);
		
		WinDrawTruncChars(msg, StrLen(msg),
						  x + kAttnListTextOffset,
						  y, 
						  w - kAttnListTextOffset);
		if( mesg_data )
		{
			WinDrawTruncChars(mesg_data, StrLen(mesg_data),
						  x + kAttnListTextOffset,
						  y + 10, 
						  w - kAttnListTextOffset);
		}
		FntSetFont (curFont);
	}
	else if ( kAttnCommandGoThere == command )
	{
		// Clear notifications
		IMResetNotifications();

		// Launch IM client
		FtrGet(CREATOR,MUNDU_CARDNO_FTR,&munduCard);
		FtrGet(CREATOR,MUNDU_DBID_FTR,&munduDb);

		if( munduDb )
		{
                	SysUIAppSwitch(munduCard, munduDb, 
				sysAppLaunchCmdNormalLaunch, NULL);
		}
	}
	else if ( kAttnCommandGotIt == command )
	{
		// Reset notification count to 0
		FtrSet(CREATOR,IM_COUNT_FTR,0);

		// Erase stored user / message
		if( user_data ) 
		{
			FtrPtrFree(CREATOR,IM_USER_PTR_FTR);
			FtrSet(CREATOR,IM_USER_PTR_FTR,NULL);
		}

		if( mesg_data ) 
		{
			FtrPtrFree(CREATOR,IM_MESG_PTR_FTR);
			FtrSet(CREATOR,IM_MESG_PTR_FTR,NULL);
		}
	}

	return errNone;
}

void IMInitNotifications(void)
{	
	// Reset notification count
	FtrSet(CREATOR,IM_COUNT_FTR,0);
}

void IMResetNotifications(void)
{
	UInt32 cardNo;
	LocalID dbID;
	Char *t;
	UInt32 i;

	// Check if there's any outstanding notifications. If not, return.
	FtrGet(CREATOR,IM_COUNT_FTR,&i);
	if( i == 0 )
		return;

	// Reset count to 0
	FtrSet(CREATOR,IM_COUNT_FTR,0);

	// Erase stored user / message
	t = NULL;
	FtrGet(CREATOR,IM_USER_PTR_FTR,&t);
	if( t ) 
	{
		FtrPtrFree(CREATOR,IM_USER_PTR_FTR);
		FtrSet(CREATOR,IM_USER_PTR_FTR,NULL);
	}

	t = NULL;
	FtrGet(CREATOR,IM_MESG_PTR_FTR,&t);
	if( t ) 
	{
		FtrPtrFree(CREATOR,IM_MESG_PTR_FTR);
		FtrSet(CREATOR,IM_MESG_PTR_FTR,NULL);
	}

	// Tell Attention Manager to resolve any outstanding notification
	FtrGet(CREATOR,MY_CARDNO_FTR,&cardNo);
	FtrGet(CREATOR,MY_DBID_FTR,&dbID);
	AttnForgetIt(cardNo, dbID, 0);

}

void IMAddNotification(Char *user, Char *mesg)
{
	UInt32 cardNo;
	LocalID dbID;

	void *attn_ptr; 

	char system[13]="aolsystemmsg";

	Char *user_data = NULL;
	Char *mesg_data = NULL;
	UInt32 counter;

	// Get my database ID
	FtrGet(CREATOR,MY_CARDNO_FTR,&cardNo);
	FtrGet(CREATOR,MY_DBID_FTR,&dbID);

	// If we're AOL System Msg, don't bother setting an attention 
	if( !strcmp(user,system) ) 
	{
		return;
	}

	// Get / update counter
	FtrGet(CREATOR,IM_COUNT_FTR,&counter);
	counter++;
	FtrSet(CREATOR,IM_COUNT_FTR,counter);

	// Get message data
	FtrGet(CREATOR,IM_USER_PTR_FTR,&user_data);
	FtrGet(CREATOR,IM_MESG_PTR_FTR,&mesg_data);

	// Get pointer to attention handler
	attn_ptr = NULL;
	FtrGet(CREATOR, FTR_ATTN_PTR, &attn_ptr);

	// First message => save data
	if( counter == 1 )
	{
		// Save the information
		FtrPtrNew(CREATOR, IM_USER_PTR_FTR, StrLen(user)+1, &user_data);
		DmWrite(user_data, 0, user, StrLen(user)+1);

		FtrPtrNew(CREATOR, IM_MESG_PTR_FTR, StrLen(mesg)+1, &mesg_data);
		DmWrite(mesg_data, 0, mesg, StrLen(mesg)+1);

		// Trigger Attention Manager
		AttnGetAttention(cardNo, dbID, 
			0,
			attn_ptr, 
			kAttnLevelSubtle, kAttnFlagsLEDBit,
			0, 0);	

	}
	else
	{
		// Subsequent => delete message, 
		if( mesg_data )
		{
			FtrPtrFree(CREATOR,IM_MESG_PTR_FTR);
			FtrSet(CREATOR,IM_MESG_PTR_FTR,NULL);
		}
	
		// And delete user if it doesn't match
		if( user_data && strcmp(user_data, user) )
		{
			FtrPtrFree(CREATOR,IM_USER_PTR_FTR);
			FtrSet(CREATOR,IM_USER_PTR_FTR,NULL);
		}

		// Tell Attention Manager to redraw
		AttnUpdate(cardNo, dbID, 0, attn_ptr, NULL, NULL, NULL);
	}
}

void IMDrawIcon(int x, int y)
{
	DmOpenRef dbref;
	MemHandle resH;
	UInt32 cardNo, dbID;

	FtrGet(CREATOR,MUNDU_CARDNO_FTR,&cardNo);
	FtrGet(CREATOR,MUNDU_DBID_FTR,&dbID);

	dbref = DmOpenDatabase(cardNo,dbID,dmModeReadOnly);

	if( dbref )
		resH = DmGet1Resource(bitmapRsc, 
			MUNDU_CHAT_BITMAP);
	else
	{
		resH = 0;
	}

	if( resH ) 
	{
		WinDrawBitmap((BitmapType*)MemHandleLock(resH), x, y);
		MemHandleUnlock(resH);
		DmReleaseResource(resH);
	}

	if( dbref ) 
		DmCloseDatabase(dbref);
}

static void RegisterForNotifications()
{
	MemHandle hndl;

	void *memptr;
	void *ftrptr;
	void *hn_ptr;
	void *ptr1;
	char *ptr2, *ptr3;

	UInt16 cardNo;
	LocalID dbID;

	if( 0 != SysCurAppDatabase(&cardNo, &dbID) )
		return;

	hndl = DmGet1Resource('code',1);
	memptr = MemHandleLock(hndl);

	FtrPtrNew(CREATOR, FTR_PTR, MemHandleSize(hndl), &ftrptr);
	DmWrite(ftrptr, 0, memptr, MemHandleSize(hndl));
	MemHandleUnlock(hndl);

	// Compute location of HandleNotification
	hn_ptr = ((void *)ftrptr)+STARTUP_CODE_SIZE;

	DmDatabaseProtect(cardNo, dbID, true);

	SysNotifyRegister(cardNo, dbID, sysNotifyEventDequeuedEvent, hn_ptr, 
			0, NULL);

 	SysNotifyRegister(cardNo, dbID, sysNotifyDeleteProtectedEvent, NULL,
			0, NULL);

	// Compute and save location of HandleAttentionManager
	ptr1 = hn_ptr;
	ptr2 = &HandleAttentionMgr;
	ptr3 = &HandleNotification;
	ptr1 += ptr2 - ptr3;
	FtrSet(CREATOR, FTR_ATTN_PTR, ptr1);

}

static void DeregisterNotifications()
{	
	Int32 val;
	void *ftrptr;

	UInt16 cardNo;
	LocalID dbID;

	val = 0; FtrGet(CREATOR, FTR_PTR, &val); ftrptr = (void *)val;

	// Check if we're registered
	if( ftrptr == NULL )
		return; 

	if( 0 != SysCurAppDatabase(&cardNo, &dbID) )
		return;

	SysNotifyUnregister(cardNo, dbID, sysNotifyEventDequeuedEvent,
			0);

 	SysNotifyUnregister(cardNo, dbID, sysNotifyDeleteProtectedEvent,
			0);

	// release feature memory
	FtrPtrFree(CREATOR, FTR_PTR);
	DmDatabaseProtect(cardNo, dbID, false);
}


//// PilotMain /////////////////////////////////////////////////////////
// It all starts here...

UInt32 PilotMain(UInt16 launchCode, MemPtr cmdPBP, UInt16 launchFlags) 
{
	Err err = 0;
	EventType event;
	UInt16 cardNo; 
	LocalID dbID;
        DmSearchStateType sstate;

	switch( launchCode ) 
	{
		case sysAppLaunchCmdSystemReset :
			// Identify me
			SysCurAppDatabase( &cardNo, &dbID );
			FtrSet(CREATOR,MY_CARDNO_FTR, cardNo);
			FtrSet(CREATOR,MY_DBID_FTR, dbID);

			// Identify MunduIM
			cardNo = 0;
			dbID = 0;

		        DmGetNextDatabaseByTypeCreator(true,&sstate,
				'appl','GMV4',true,&cardNo, &dbID);

			// If not found, don't install
			if( (cardNo == 0) && (dbID == 0) )
				break;

        		FtrSet(CREATOR,MUNDU_CARDNO_FTR,cardNo);
        		FtrSet(CREATOR,MUNDU_DBID_FTR,dbID);

			// Register for notifications
			RegisterForNotifications();
			break;

		case sysAppLaunchCmdNotify : 
			HandleNotification((SysNotifyParamType *)cmdPBP);
			break;

		case sysAppLaunchCmdAttention :
		{
			/*
			 * get parameters
			 */
			AttnLaunchCodeArgsType *launchArgsP = 
					(AttnLaunchCodeArgsType*)cmdPBP;
			AttnCommandType command = launchArgsP->command;
			AttnCommandArgsType *commandArgsP = launchArgsP->commandArgsP;

			HandleAttentionMgr(command, 
					 launchArgsP->userData,
					 commandArgsP);

			break;
		}

		case sysAppLaunchCmdNormalLaunch :
		        DmGetNextDatabaseByTypeCreator(true,&sstate,
				'appl','GMV4',true,&cardNo, &dbID);

			IMAddNotification("test user","test message");

			// Well, uh, we don't have a UI, so we just quit
			MemSet( &event, sizeof(event), 0);

			event.eType = appStopEvent;
			EvtAddEventToQueue(&event);
			break;

		default:
			break;
	}


	return err;
}

